#ifdef NDEBUG
#  undef NDEBUG // force assert() to work
#endif
#include <cassert>
#include <memory> // auto_ptr

#include <stdlib.h> // getenv()

#include <s11n.net/s11n/s11nlite.hpp>
#include <s11n.net/s11n/simple_config.hpp>
// #include <s11n.net/s11n/proxy/std/list.hpp>
#include <s11n.net/s11n/s11n_debuggering_macros.hpp>
// #include <s11n.net/acme/argv_parser.hpp>
// #include <s11n.net/cl/dll_util.hpp>
#include <s11n.net/s11n/nodeutil/node_tree.hpp>

#include "argv_parser.hpp"


#include <qt.h>
#include <qsize.h>
#include <qpoint.h>

#include "s11nbrowser_config.hpp"
#include "S11nNodeListViewItem.h"
#include "s11nbrowserwindow.h" // Qt-generated file


#include "qt_serialize.hpp"

typedef s11n::util::node_tree<s11nlite::node_type> TreeUtil;
TreeUtil & node_tree()
{
        static TreeUtil bob;
        return bob;
}

s11nlite::simple_config & config()
{
	static s11nlite::simple_config bob("s11nbrowser");
	return bob;
}

QListView * m_list = 0;
QListView & node_list()
{
        return *m_list;
}

s11nlite::node_type * m_root = 0;
s11nlite::node_type * root_node()
{
        return m_root;
}

/**
   Deletes previous root_node(), clears node_list() and node_tree(),
   and sets new root to n. Should return the total number of nodes
   under n, including n.
*/
size_t root_node( s11nlite::node_type * n )
{
        if( n != m_root)
        {
                delete m_root;
                m_root = n;
        }
        node_tree().clear();
        node_list().clear();
        size_t c = 0;
        if( n )
        {
                c = node_tree().climb_tree( n );
                typedef s11n::qt::S11nNodeListViewItem NLI;
                new NLI( *n, &node_list() ); // node_list() owns the new NLI.
                node_list().firstChild()->setOpen( true );
        }
        return c;
}


bool steal_node( s11nlite::node_type * node )
{
        if( ! node ) return false;
        return node_tree().take_node( node );
//        return 0 != node_tree().climb_tree( root_node() );
}

s11nlite::node_type * load_node_file( const std::string & filename,
                                      const std::string & serclass )
{

        typedef s11nlite::node_type NT;
        
        typedef s11nlite::serializer_interface SER;
        typedef std::auto_ptr<SER> APS;

        NT * node = 0;
        if( serclass.empty() )
        {
                node = s11nlite::load_node( filename );
        }
        else
        {
                APS ser = APS( s11nlite::create_serializer( serclass ) );
                if( ! ser.get() )
                {
                        CERR << "load_node_file("<<filename<<","<<serclass<<"): couldn't classload requested Serializer type.\n";
                        return false;
                }
                node = ser->deserialize( filename );
        }
        if( ! node  )
        {
                CERR << "load_node_file("<<filename<<","<<serclass<<") failed to load a node :(\n";
                return 0;
        }
//         size_t count = 
//         node_tree().climb_tree( root_node() );
//         CERR << count << " node"<<(count==1 ? "" : "s")<<" read in.\n";

//          typedef s11n::qt::S11nNodeListViewItem NLI;
//         new NLI( * root_node(), &node_list() ); // node_list() owns the new NLI.
        return node;
}

struct root_node_cleaner
{
        root_node_cleaner() {}
        ~root_node_cleaner()
        {
                delete root_node();
        }

};


/**
   Loads Serializer DLLs.
*/
// void autoload_serializers()
// {
//         typedef std::list<std::string> SList;
//         SList li;
// #define SER(X) li.push_back( "s11n/" # X "_serializer" );
//         SER(mysql);
// #undef SER

//         SList::iterator it = li.begin(), et = li.end();
//         for( ; et != it; ++it )
//         {
//                 try
//                 {
//                         ::s11n::plugin::open_dll( *it );
//                 }
//                 catch( const cllite::cl_exception & ex )
//                 {
//                         CERR << "Autoload of '"<<(*it)<<"' failed. :(\n"
//                              << ex.what() << "\n";
//                         continue;
//                 }
//         }
// }

void show_help()
{
	std::cout << s11n_BROWSER_NAME << " Version " << s11n_BROWSER_VERSION
		  << "\nLicense: " << s11n_BROWSER_LICENSE_TEXT
		  << "\nCommand-line options:\n\n"
		  << "--help, -?:\n\tShow help text and exit with code 0.\n"
		  << "-dl DLLNAME:\n\tOpens a DLL at startup and exits if it cannot be found/opened.\n"
		  << "-DL DLLNAME:\n\tLike -dl, but does not exit if it fails.\n"
		  << "-f FILENAME:\n\tLoads FILENAME (an s11n file) at startup.\n"
		  << "-s Serializer:\n\tSets the preferred output format.\n"
		  << "\n";
	
}

// config file keys:
#define CFGKEY_SIZE "window_size"
#define CFGKEY_SERIALIZER "serializer_class"

int
load_dlls( int argc, char **argv )
{
	if( argc < 2 ) return 0;
	std::string arg;
	std::string foo;
	int mode = 0; // 1 == tolerant, 2 == intolerant
	for( int i = 1; i < argc; ++i )
	{
		mode = 0;
		arg = argv[i];
		if( "-DL" == arg ) mode = 1; // tolerant
		else if( "-dl" == arg ) mode = 2; // intolerant
		else continue;

#if ! s11n_CONFIG_ENABLE_PLUGINS
		throw std::runtime_error( "-dl/-DL are not supported because your libs11n was built without plugins support." );
#else
		if( argc == ++i )
		{
			throw std::runtime_error( "-dl/-DL parameters require a following argument (filename or DLL class name)" );
		}
		arg = argv[i];
		foo = s11n::plugin::open( arg );
		if( foo.empty() )
		{
			std::ostringstream errmsg;
			errmsg << "DLL error: '"<<arg<<"': " << s11n::plugin::dll_error();
			if( 1 == mode )
			{
				CERR << "Warning: " << errmsg.str() << '\n';
				continue;
			}
			throw std::runtime_error( errmsg.str() );
		}
#endif // s11n_CONFIG_ENABLE_PLUGINS
	}
	return 0;
}

int main( int argc, char ** argv )
{
	s11n::Private::s11n_1_1_assertion(); // see the API docs for why this is here!

        enum {
        ErrorDLL = 1
        };
	

        typedef s11nlite::node_traits TR;

	cliutil::argv_map args;
	cliutil::unnamed_args_list unused;

	s11nlite::node_type argn; // we're gonna cheat a bit and use this for arg parsing
	
	cliutil::parse_args( argc, argv, 1, args, unused );
	
	cliutil::argv_map::const_iterator amit = args.begin();
	for( ; args.end() != amit; ++amit )
	{
		TR::set( argn, (*amit).first, (*amit).second );
	}
	
//         s11nlite::node_type cfg;


	try
	{
		load_dlls( argc, argv );
	}
	catch( std::exception const & ex )
	{
		CERR << "EXCEPTION: " << ex.what() << '\n';
		return ErrorDLL;
	}

        root_node_cleaner cleaner; // root will be cleaned when this obj goes out of scope

        /////////////////////////////////////////////////////////////////////
        // load requested DLLs
//         typedef std::list<cllite::dll::dlentry> DList;
//         DList dlllist;
//         if( 0 != cllite::dll::parse_dlentry_flags( argc -1 , argv+1, dlllist ) )
//         {
//                 try
//                 {
//                         cllite::dll::dlload( dlllist );
//                 }
//                 catch( const cllite::cl_exception & ex )
//                 {
//                         CERR << ex.what() << "\n";
//                         return ErrorDLL;
//                 }
//                 if( 0 )
//                 {
//                         CERR << "dlentries:\n";
//                         s11nlite::save( dlllist, std::cout );
//                         DList delist;
//                         s11nlite::s11n_cast( dlllist, delist );
//                         s11nlite::save( delist, std::cout );
//                 }
//         }
//         autoload_serializers();
	// end DLL loading.
	////////////////////////////////////////////////////////////////////////


//         args.set_help( "f FILENAME", "Loads the given file on startup." );
//         args.set_help( "?, help", "Show this help." );
//         args.set_help( "dl dllname",
//                        std::string("Opens the given DLL. ")
//                        +"May be used multiple times. All DLLs are loaded before the Serializer (-s) is loaded, "
//                        +"DLL names may be absolute file names or base filenames if those files are in the lib path. "
//                        +"This is useful when, e.g., the Serializer treats filenames as db keys or something similar."
//                        +" Use -v to see information about what DLLs get loaded this way."
//                        );

//         args.set_help( "DL dllname",
//                        std::string("Exacly like -dl")+
//                        ", but does not abort if the dll is not found or if there is an error opening it. "+
//                        "In s11n slang this is called a \"tolerant dll\"." );

//         args.set_help( "s", "Sets the default s11nlite Serializer (i.e., the default output format)." );

	if( TR::is_set(argn,std::string("?")) || TR::is_set(argn,std::string("help")) )
         {
		 show_help();
                 return 0;
         }

	std::string serclass = TR::get(argn,std::string("s"), TR::get(config().node(), CFGKEY_SERIALIZER ,std::string("")) );
         if( ! serclass.empty() )
         {
                 s11nlite::serializer_class( serclass );
         }





        QApplication app( argc, argv );

// When using a db i get this bogus error from qt:
//     QSqlDatabase: QMYSQL3 driver not loaded
//     QSqlDatabase: available drivers: QMYSQL3 QPSQL7
// But the db access works fine via Qt Designer.
//
// #define DB_MYS11N_DRIVER     "QMYSQL3"
// #define DB_MYS11N_DBNAME     "mys11n"
// #define DB_MYS11N_USER       ""
// #define DB_MYS11N_PASSWD     ""
// #define DB_MYS11N_HOST       ""
//         QSqlDatabase *defaultDB = QSqlDatabase::addDatabase( DB_MYS11N_DRIVER );
//         defaultDB->setDatabaseName( DB_MYS11N_DBNAME );
//         defaultDB->setUserName( DB_MYS11N_USER );
//         defaultDB->setPassword( DB_MYS11N_PASSWD );
//         defaultDB->setHostName( DB_MYS11N_HOST );
//         if ( ! defaultDB->open() ) {
//                 CERR << "Db open failed :(\n";
//         }


        S11nBrowserWindow * w = new S11nBrowserWindow;
        w->setCaption( QString("s11n Data Node Browser ver. ") + s11n_BROWSER_VERSION );
        app.setMainWidget( w );
        m_list = w->listNodes;
        QSize wsz = TR::get( config().node(), CFGKEY_SIZE, QSize(600,500) );
        w->resize( wsz );
        w->show();

	std::string fname = TR::get(argn,std::string("f"), std::string() );
        if( !  fname.empty() )
        {
                 size_t count = root_node( load_node_file( fname, "" ) );
                 CERR << count << " node" << (count==1 ? "" : "s") << " loaded from "<<fname<<".\n";
        }


        int ret = app.exec();
        wsz = w->size();
        TR::set( config().node(), CFGKEY_SIZE, wsz  ); // i'm surprised that w is still valid here.
	TR::set( config().node(), CFGKEY_SERIALIZER, s11nlite::serializer_class() );
        root_node( 0 );
        delete( w );
        return ret;
}

